// decodable-nnet-online-for-hark.h
// If you want to put this file under the directory, it file is placed in under the "nnet" directory.
// Implements like "nnet2/online-nnet2-decodable.h".
// Implements like "nnet3/online-nnet3-decodable.h".

#ifndef KALDI_DECODABLE_NNET1_ONLINE_FOR_HARK_H_
#define KALDI_DECODABLE_NNET1_ONLINE_FOR_HARK_H_

#include "itf/online-feature-itf.h"
#include "itf/decodable-itf.h"
#include "hmm/transition-model.h"
// nnet1
#include "nnet/nnet-nnet.h"
#include "nnet/nnet-pdf-prior.h"

namespace kaldi {
namespace nnet1 {

struct DecodableNnetOnlineForHarkOptions {
  kaldi::BaseFloat acoustic_scale;
  kaldi::int32 max_nnet_batch_size;

  DecodableNnetOnlineForHarkOptions():
      acoustic_scale(0.1),
      max_nnet_batch_size(128/*256*/) { }

  // for kaldi compatibility
  void Register(OptionsItf *opts) {
    opts->Register("acoustic-scale", &acoustic_scale,
                   "Scaling factor for acoustic likelihoods");
    opts->Register("max-nnet-batch-size", &max_nnet_batch_size,
                   "Maximum batch size we use in neural-network decodable object, "
                   "in cases where we are not constrained by currently available "
                   "frames (this will rarely make a difference)");
  }
  // for avoid of option name conflict in multi nnet type support
  void SetAcousticScale(kaldi::BaseFloat arg) {
    acoustic_scale = arg;
  }
  void SetMaxNnetBatchSize(kaldi::int32 arg) {
    max_nnet_batch_size = arg;
  }
};

class DecodableNnetOnlineForHark : public DecodableInterface {
 public:
  DecodableNnetOnlineForHark(const kaldi::nnet1::Nnet &nnet,
                             kaldi::nnet1::PdfPrior &pdf_prior,
                             const kaldi::TransitionModel &trans_model,
                             const DecodableNnetOnlineForHarkOptions &opts,
//                             kaldi::BaseFloat ac_scale,
                             //kaldi::OnlineFeatureInterface *features_if);
                             std::shared_ptr<kaldi::OnlineFeatureInterface> features_if);

  /// Returns the scaled log likelihood
  virtual kaldi::BaseFloat LogLikelihood(kaldi::int32 frame, kaldi::int32 index);

  virtual bool IsLastFrame(kaldi::int32 frame) const;

  virtual kaldi::int32 NumFramesReady() const;

  /// Indices are one-based!  This is for compatibility with OpenFst.
  virtual kaldi::int32 NumIndices() const { return trans_model_.NumTransitionIds(); }

 private:
  void ComputeForFrame(kaldi::int32 frame);

  //kaldi::OnlineFeatureInterface *features_if_;
  std::shared_ptr<kaldi::OnlineFeatureInterface> features_if_;
  kaldi::nnet1::Nnet nnet_;
  kaldi::nnet1::PdfPrior &pdf_prior_;  // log-priors taken from the model.
  const kaldi::TransitionModel &trans_model_;
  DecodableNnetOnlineForHarkOptions opts_;
//  kaldi::BaseFloat ac_scale_;
//  kaldi::int32 max_nnet_batch_size_;
  kaldi::int32 feat_dim_;
  kaldi::int32 num_pdfs_;
  kaldi::int32 begin_frame_;

  kaldi::Matrix<kaldi::BaseFloat> scaled_loglikes_;

  //KALDI_DISALLOW_COPY_AND_ASSIGN(DecodableNnetOnline);
};

} // namespace nnet1
} // namespace kaldi

#endif // KALDI_DECODABLE_NNET1_ONLINE_FOR_HARK_H_
