#ifdef _MSC_VER
#define NOMINMAX
#endif

#undef ENABLE_SPEECHRECOG_COMPATIBILITY

#include <thread>

#include <time.h>

#include "mp/stdio_synchronizer.hpp"
#include "mp/sleep_wrapper.hpp"
#include "mp/time_wrapper.hpp"
#include "mp/file_wrapper.hpp"
#include "mp/tcp_client.hpp"
#include "mp/tcp_server.hpp"
//#include "mp/mfcnet_server.hpp"
//#include "mp/result_server.hpp"

#include "print_license.hpp"

#include <signal.h>
#include <vector>
#include "base/kaldi-common.h"
#include "base/kaldi-error.h"
#include "util/common-utils.h"

//#include "tree/context-dep.h"
//#include "hmm/transition-model.h"

#include "matrix/kaldi-matrix.h"
#include "feat/online-feature.h"
#include "ivector/ivector-extractor.h"
#include "decoder/lattice-faster-online-decoder.h"

//#include "decoder/decoder-wrappers.h"
//#include "nnet2/decodable-am-nnet.h"
//#include "base/timer.h"

#include "fstext/kaldi-fst-io.h"
#include "lat/lattice-functions.h"
#include "nnet3/nnet-utils.h"
#include "online2/online-nnet2-feature-pipeline.h"
#include "online2/online-ivector-feature.h"
#include "online-nnet2-feature-pipeline-mod.h"

#include "nnet3/decodable-simple-looped.h"

//#include "thread/kaldi-task-sequence.h"

#include "decodable-nnet1-online-for-hark.h"
#include "decodable-nnet2-online-for-hark.h"
#include "decodable-nnet3-online-for-hark.h"
#include "online-mfcnet.h"
#include "recognizer0.h"
#include "recognizer1.h"
#include "recognizer2.h"
#include "recognizer3.h"

//#include "pthread.h"

//using namespace std;
//using namespace kaldi;
//using namespace kaldi::nnet1;
//using namespace fst;
//using kaldi::int32;

bool is_active = true;

void signal_handler_int(int sig_num)
{
  is_active = false;
  _raw_print_with_endl("There accepted termination by keyboard interrupt.");
  _raw_print_with_endl("Please wait until the currently processing child thread(s) ends.");
}

namespace kaldi {
  // We pass print message received from kaldi message handler to the stdio synchronizer.
  void our_log_handler(const kaldi::LogMessageEnvelope &envelope, const char *message){
    if(envelope.severity == kaldi::LogMessageEnvelope::kInfo){         _info_print(std::string(message));    return; }
    if(envelope.severity == kaldi::LogMessageEnvelope::kWarning){      _warning_print(std::string(message)); return; }
    if(envelope.severity == kaldi::LogMessageEnvelope::kError){        _error_print(std::string(message));   return; }
    if(envelope.severity == kaldi::LogMessageEnvelope::kAssertFailed){ _error_throw(std::string(message));   return; }
  }
}

void nnet1_recog_proc_thread(kaldi::nnet1::NnetRecognizer* recognizer)
{
#if HAVE_CUDA == 1
  // Setting the device id for each new thread
  // Reference: http://devblogs.nvidia.com/parallelforall/cuda-pro-tip-always-set-current-device-avoid-multithreading-bugs/
  if(kaldi::CuDevice::Instantiate().Enabled()){
    int active_gpu_id = kaldi::CuDevice::Instantiate().ActiveGpuId();
    cudaError_t e = cudaSetDevice(active_gpu_id);
    if(e != cudaSuccess){ KALDI_ERR << "Failed to set device " << active_gpu_id << ": " << cudaGetErrorString(e); }
    KALDI_LOG << "running thread on device " << active_gpu_id;
  }
#endif
  recognizer->Recognize();
  delete recognizer;
  std::stringstream ss;
  ss << "recognition end " << multiplatform::time::formatted_time() << std::endl;
  ss << "terminate thread(" << std::this_thread::get_id() << ")";
  _raw_print_with_endl(ss.str());
}

void nnet2_recog_proc_thread(kaldi::nnet2::NnetRecognizer* recognizer)
{
#if HAVE_CUDA == 1
  // Setting the device id for each new thread
  // Reference: http://devblogs.nvidia.com/parallelforall/cuda-pro-tip-always-set-current-device-avoid-multithreading-bugs/
  if(kaldi::CuDevice::Instantiate().Enabled()){
    int active_gpu_id = kaldi::CuDevice::Instantiate().ActiveGpuId();
    cudaError_t e = cudaSetDevice(active_gpu_id);
    if(e != cudaSuccess){ KALDI_ERR << "Failed to set device " << active_gpu_id << ": " << cudaGetErrorString(e); }
    KALDI_LOG << "running thread on device " << active_gpu_id;
  }
#endif
  recognizer->Recognize();
  delete recognizer;
  std::stringstream ss;
  ss << "recognition end " << multiplatform::time::formatted_time() << std::endl;
  ss << "terminate thread(" << std::this_thread::get_id() << ")";
  _raw_print_with_endl(ss.str());
}

void nnet3_recog_proc_thread(kaldi::nnet3::NnetRecognizer* recognizer)
{
#if HAVE_CUDA == 1
  // Setting the device id for each new thread
  // Reference: http://devblogs.nvidia.com/parallelforall/cuda-pro-tip-always-set-current-device-avoid-multithreading-bugs/
  if(kaldi::CuDevice::Instantiate().Enabled()){
    int active_gpu_id = kaldi::CuDevice::Instantiate().ActiveGpuId();
    cudaError_t e = cudaSetDevice(active_gpu_id);
    if(e != cudaSuccess){ KALDI_ERR << "Failed to set device " << active_gpu_id << ": " << cudaGetErrorString(e); }
    KALDI_LOG << "running thread on device " << active_gpu_id;
  }
#endif
  recognizer->Recognize();
  delete recognizer;
  std::stringstream ss;
  ss << "recognition end " << multiplatform::time::formatted_time() << std::endl;
  ss << "terminate thread(" << std::this_thread::get_id() << ")";
  _raw_print_with_endl(ss.str());
}

inline void print_license_notice()
{
  const char* license_notice =
    "\n"
    "  This software includes work that is distributed under the Apache License 2.0 .\n"
    "\n"
    "  This software uses the following libraries.\n"
    "  [Apache License 2.0 libraries]\n"
    "    - Kaldi (http://kaldi-asr.org/)\n"
    "    - OpenFST (http://www.openfst.org/)\n"
    "  [3-clause BSD libraries]\n"
    "    - OpenBLAS (https://github.com/xianyi/OpenBLAS)\n"
    "\n"
    "  The license of the dependencies can be confirmed with the following commands.\n"
    "    kaldidecoder --print-kaldi-license\n"
    "    kaldidecoder --print-openfst-license\n"
    "    kaldidecoder --print-openblas-license\n"
    "\n";
  _info_print(license_notice);
}

int main(int argc, char* argv[])
{
  // std::make_unique() method will be support from the C++14 so I made a mistake. I recovered old code.
  //multiplatform::default_synchronizer = std::make_unique<multiplatform::stdio_synchronizer>();
  multiplatform::default_synchronizer.reset(new multiplatform::stdio_synchronizer());
  print_license_notice();

  // We take over the output of kaldi with our stdio synchronizer class.
  kaldi::SetLogHandler(kaldi::our_log_handler);

  // usage string
  const char* usage =
    "usage: If you requests need use ONLINE decoding with nnet1 model.\n"
    "       (ONLINE mode is default)\n"
    "  e.g.) kaldidecoder [--port-mfcnet=5530]"
                        " [--port-result=10500]"
                        " --max-active=2000"
                        " --filename-words=words.txt"
                        " --filename-align-lexicon=align_lexicon.int"
                        " --filename-feature-transform=final.feature_transform"
                        " --filename-nnet=final.nnet"
                        " --filename-mdl=final.mdl"
                        " --filename-class-frame-counts=ali_train_pdf.counts"
                        " --filename-fst=HCLG.fst\n"
    "note: This online mode to simulate behave of the julius_mft (mfcnet) module mode.\n"
    "\n"
    "usage: If you requests need use OFFLINE decoding with nnet1 model.\n"
    "       (OFFLINE mode is must need --filelist option)\n"
    "  e.g.) kaldidecoder --filename-features-list=features_list.txt"
                        " --max-active=2000"
                        " --filename-words=words.txt"
                        " --filename-align-lexicon=align_lexicon.int"
                        " --filename-feature-transform=final.feature_transform"
                        " --filename-nnet=final.nnet"
                        " --filename-mdl=final.mdl"
                        " --filename-class-frame-counts=ali_train_pdf.counts"
                        " --filename-fst=HCLG.fst\n"
    "\n"
    "usage: If you requests need use ONLINE decoding with nnet3/chain model.\n"
    "       (ONLINE mode is default)\n"
    "  e.g.) kaldidecoder [--port-mfcnet=5530]"
                        " [--port-result=10500]"
                        " --nnet-type=3"
                        " --max-active=2000"
                        " --filename-words=words.txt"
                        " --filename-align-lexicon=align_lexicon.int"
                        " --filename-mdl=final.mdl"
                        " --filename-fst=HCLG.fst\n"
    "note: This online mode to simulate behave of the julius_mft (mfcnet) module mode.\n"
    "      If you are using the iVectors you will need the following additional options.\n"
    "      '--ivector-extraction-config=ivector_extractor.conf'\n"
    "\n"
    "usage: If you requests need use OFFLINE decoding with nnet3/chain model.\n"
    "       (OFFLINE mode is must need --filelist option)\n"
    "  e.g.) kaldidecoder --filename-features-list=features_list.txt"
                        " --nnet-type=3"
                        " --max-active=2000"
                        " --filename-words=words.txt"
                        " --filename-align-lexicon=align_lexicon.int"
                        " --filename-mdl=final.mdl"
                        " --filename-fst=HCLG.fst\n"
    "note: If you are using the iVectors you will need the following additional options.\n"
    "      '--ivector-extraction-config=ivector_extractor.conf'\n";

  signal(SIGINT, signal_handler_int);

  kaldi::ParseOptions po(usage);

  kaldi::LatticeFasterDecoderConfig decoder_config;
  kaldi::OnlineNnet2FeaturePipelineConfig feature_config;
  kaldi::OnlineNnet2FeaturePipelineInfo *feature_info = nullptr;
  kaldi::nnet3::NnetSimpleLoopedComputationOptions decodable_opts;

  bool enable_debug = false;
#if HAVE_CUDA == 1
  bool enable_cuda = true;
#endif
  bool print_kaldi_license = false;
  bool print_openfst_license = false;
  bool print_openblas_license = false;

  int port_mfcnet = 5530;
  int port_result = 10500;
  int nnet_type = 1; // 1, 2 or 3

  std::string host_mfcnet = "localhost";
  std::string host_result = "localhost";

  std::string filename_features_list = "";

  std::string filename_feature_transform = "";
  std::string filename_nnet = "";
  std::string filename_mdl = "";
  std::string filename_class_frame_counts = "";
  std::string filename_fst = "";

  std::string filename_align_lexicon = "";

  std::string filename_phones = "";
  std::string filename_words = "";

  kaldi::int32 splice = 3;
  std::string lm_name = "";

  std::uint32_t max_tasks = 4; // 0 is no limit

  //bool enable_stdout = true;
  int progress_output_interval = 0;
  int nbest_count = 1;

  // function enable/disable
  po.Register("enable-debug", &enable_debug, "Enable debug function");
#if HAVE_CUDA == 1
  po.Register("enable-cuda", &enable_cuda, "Enable CUDA processing");
#endif

  // set threads
  po.Register("max-tasks", &max_tasks, "Specify the number of parallel threads. If the parameter is greater than 0, the number of threads is limited to the specified number and unlimited if it is set to 0. If you want to do offline decoding, setting to 0 will decode all features in the feature list in parallel.");

  // license notice
  po.Register("print-kaldi-license", &print_kaldi_license, "Print the license information of the Kaldi library we are using.");
  po.Register("print-openfst-license", &print_openfst_license, "Print the license information of the OpenFST library we are using.");
  po.Register("print-openblas-license", &print_openblas_license, "Print the license information of the OpenBLAS library we are using.");

  // function mode setting
  po.Register("filename-features-list", &filename_features_list, "List of feature files to recognize (file input mode only)");
  po.Register("host-mfcnet", &host_mfcnet, "host/ip of features input (mfcnet protocol) (only effective in online decoding)");
  po.Register("port-mfcnet", &port_mfcnet, "port number of features input (mfcnet protocol) (only effective in online decoding)");
  po.Register("host-result", &host_result, "host/ip of recognition result output");
  po.Register("port-result", &port_result, "port number of recognition result output");
  po.Register("nnet-type", &nnet_type, "nnet type (1: nnet1 model, 3: nnet3/chain model)");
  //po.Register("enable-stdout", &enable_stdout, "Enable output to stdout");
  po.Register("progress-output-interval", &progress_output_interval, "Set the output interval in number of frames of the intermediate results. If 0 is set, intermediate results are not output. (only effective in online decoding)");
  po.Register("nbest_count", &nbest_count, "Set the number of n-best results");

  // input file setting
  po.Register("filename-align-lexicon", &filename_align_lexicon, "[required] align lexicon file path (e.g. align_lexicon.int)");
  po.Register("filename-feature-transform", &filename_feature_transform, "[required only for nnet1] feature transform file path (e.g. final.feature_transform)");
  po.Register("filename-nnet", &filename_nnet, "[required only for nnet1] nnet file path (e.g. final.nnet)");
  po.Register("filename-mdl", &filename_mdl, "[required] model file path (e.g. final.mdl)");
  po.Register("filename-class-frame-counts", &filename_class_frame_counts, "[required only for nnet1] class frame counts file path (e.g. ali_train_pdf.counts)");
  po.Register("filename-fst", &filename_fst, "[required] HCLG FST file path (e.g. HCLG.fst)");

  po.Register("filename-phones", &filename_phones, "phones symbols list file path (e.g. phones.txt)");
  po.Register("filename-words", &filename_words, "[required] words symbols list file path (e.g. words.txt)");

  po.Register("lm-name", &lm_name, "Language model name");
  po.Register("splice", &splice, "DNN input splice (time context around current frame)");

  decoder_config.Register(&po);
  feature_config.Register(&po);
  decodable_opts.Register(&po);

  po.Read(argc, argv);
  if(po.NumArgs() != 0){
    po.PrintUsage();
    _warning_print("Correct argument was not set.\n  The correct format for an argument with a value is \"--argument=value\".\n  Please check whether it is not specified like \"--argument value\".\n");
    return 255;
  }
  if(print_kaldi_license || print_openfst_license || print_openblas_license){
    if(print_kaldi_license || print_openfst_license){
      print_apache_license();
    }
    if(print_openblas_license){
      print_bsd_3clause_license();
    }
    return 0;
  }
  {
    std::stringstream ss_othererr; // A message when a incorrect configuration argument is found.
    std::stringstream ss_required; // A message when a missing argument is found.
    std::stringstream ss_mistaken; // A message when there are arguments that does not need to be specified.
    // check required filename
    if(filename_words == ""){
      ss_required << "  You must set --filename-words\n";
    }
    if(filename_align_lexicon == ""){
      ss_required << "  You must set --filename-align-lexicon\n";
    }
    if(filename_mdl == ""){
      ss_required << "  You must set --filename-mdl\n";
    }
    if(nnet_type == 1){
      // check required filename
      if(filename_nnet == ""){
        ss_required << "  You must set --filename-nnet\n";
      }
      if(filename_feature_transform == ""){
        ss_required << "  You must set --filename-feature-transform\n";
      }
      if(filename_class_frame_counts == ""){
        ss_required << "  You must set --filename-class-frame-counts\n";
      }
      // check unsupported filename
      if(feature_config.ivector_extraction_config != ""){
        ss_mistaken << "  You can not be set --ivector-extraction-config on nnet1\n";
      }
      if(feature_config.online_pitch_config != ""){
        ss_mistaken << "  You can not be set --online-pitch-config on nnet1\n";
      }
    }
    else if(nnet_type == 2){
      ss_othererr << "  Sorry, but there is no support plan for the nnet2 model. Please migrate to nnet3 model.\n";
    }
    else if(nnet_type == 3){
      // check required filename
      // check unsupported filename
      if(filename_nnet != ""){
        ss_mistaken << "  You can not be set --filename-nnet on nnet3/chain\n";
      }
      if(filename_feature_transform != ""){
        ss_mistaken << "  You can not be set --filename-feature-transform on nnet3/chain\n";
      }
      if(filename_class_frame_counts != ""){
        ss_mistaken << "  You can not be set --filename-class-frame-counts on nnet3/chain\n";
      }
    }
    else{
      ss_othererr << "  Unknown nnet type " << nnet_type << "\n";
    }
    if(!ss_othererr.str().empty()){
      std::string err_msg("A value that can not be specified is set. Please confirm usage.\n");
      po.PrintUsage();
      _error_print(err_msg.append(ss_othererr.str()));
      return 255;
    }
    if(!ss_required.str().empty()){
      std::string err_msg("The following required arguments are not set. You must set the following.\n");
      po.PrintUsage();
      _error_print(err_msg.append(ss_required.str()));
      return 255;
    }
    if(!ss_mistaken.str().empty()){
      std::string err_msg("Invalid settings are included. Please check the following settings.\n");
      po.PrintUsage();
      _warning_print(err_msg.append(ss_mistaken.str()));
      return 255;
    }
  }
  po.PrintConfig(std::cout);

// none optional arguments
//  std::string arg1 = po.GetArg(1);
//  std::string arg2 = po.GetArg(2);
//  std::string arg3 = po.GetArg(3);
//  std::string arg4 = po.GetArg(4);
//  std::string arg5 = po.GetArg(5);
//  std::string arg6 = po.GetArg(6);
//  std::string arg7 = po.GetArg(7);
//  std::string arg8 = po.GetArg(8);
//  std::string arg9 = po.GetArg(9);

#if HAVE_CUDA == 1
  if(enable_cuda){
    kaldi::CuDevice::Instantiate().SelectGpuId("yes");
    kaldi::CuDevice::Instantiate().ResetProfile();
  }
#endif

  { // common scope start
    multiplatform::wsa_wrapper wsa;
    std::shared_ptr<mfcnet_server> ms;
    std::shared_ptr<result_server> rs;
    if(!host_mfcnet.empty() && filename_features_list.empty()){
      ms = std::shared_ptr<mfcnet_server>(new mfcnet_server(host_mfcnet, port_mfcnet));
    }
    if(!host_result.empty()){
      rs = std::shared_ptr<result_server>(new result_server(host_result, port_result));
    }

    ////////////////////////////////////////////////////////////
    //////////           Word Symbol Table            //////////
    ////////////////////////////////////////////////////////////
    fst::SymbolTable *word_sym_table = nullptr;
    if(!filename_words.empty()){
      if(!(word_sym_table = fst::SymbolTable::ReadText(filename_words))){
        KALDI_ERR << "Could not read symbol table from file " << filename_words;
      }
      KALDI_LOG << "filename_words: " << filename_words;
      if(enable_debug){
        int i = 0, num_symbols = word_sym_table->NumSymbols();
        KALDI_LOG << "word_sym_table (num_symbols): " << num_symbols;
        while(i < num_symbols){ KALDI_LOG << i << " " << word_sym_table->Find(i); i++; }
      }
    }
    if(word_sym_table == NULL){
      KALDI_ERR << "Could not read symbol table from file " << filename_words;
    }

    ////////////////////////////////////////////////////////////
    //////////           Phone Symbol Table           //////////
    ////////////////////////////////////////////////////////////
    fst::SymbolTable *phone_sym_table = nullptr;
    if(!filename_phones.empty()){
      if(!(phone_sym_table = fst::SymbolTable::ReadText(filename_phones))){
        KALDI_ERR << "Could not read symbol table from file " << filename_phones;
      }
      KALDI_LOG << "filename_phones: " << filename_phones;
      if(enable_debug){
        int i = 0, num_symbols = phone_sym_table->NumSymbols();
        KALDI_LOG << "phone_sym_table (num_symbols): " << num_symbols;
        while(i < num_symbols){ KALDI_LOG << i << " " << phone_sym_table->Find(i); i++; }
      }
    }

    ////////////////////////////////////////////////////////////
    //////////          FST (Language Model)          //////////
    ////////////////////////////////////////////////////////////
    //fst::Fst<fst::StdArc> *decode_fst = fst::ReadFstKaldiGeneric(filename_fst); //new kaldi method
    //fst::Fst<fst::StdArc> *decode_fst = fst::ReadFstKaldi(filename_fst);
    // changed implements for support check FST type function and ConstFst. // ReadDecodeGraph()
    fst::Fst<fst::StdArc> *decode_fst = nullptr;
    {
      if(filename_fst.empty()){
        KALDI_WARN << "If you not set FST file, it waiting input from stdin.";
        filename_fst = "-";
      } // If filename is not specified, change to stdin.
      kaldi::Input ki(filename_fst);
      if(!ki.Stream().good()){
        KALDI_ERR << "Could not open decoding-graph FST \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      fst::FstHeader hdr;
      if(!hdr.Read(ki.Stream(), "<unknown>")){
        KALDI_ERR << "Reading FST: error reading FST header from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
      if(hdr.ArcType() != fst::StdArc::Type()){
        KALDI_ERR << "FST with arc type " << hdr.ArcType() << " not supported.";
      }
      fst::FstReadOptions ropts("<unspecified>", &hdr);
      if(hdr.FstType() == "vector"){
        decode_fst = fst::VectorFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else if(hdr.FstType() == "const"){
        decode_fst = fst::ConstFst<fst::StdArc>::Read(ki.Stream(), ropts);
      }
      else{
        KALDI_ERR << "Reading FST: unsupported FST type: " << hdr.FstType();
      }
      if(decode_fst == NULL){
        KALDI_ERR << "Could not read FST (after reading header) from \"" << kaldi::PrintableRxfilename(filename_fst) << "\".";
      }
    }

    ////////////////////////////////////////////////////////////
    //////////     Align Lexicon (Language Model)     //////////
    ////////////////////////////////////////////////////////////
    if(filename_align_lexicon.empty()){
      KALDI_ERR << "align lexicon is not specified";
    }
    std::vector<std::vector<kaldi::int32> > lexicon;
    {
      bool binary_in;
      kaldi::Input ki(filename_align_lexicon, &binary_in);
      KALDI_ASSERT(!binary_in && "Not expecting binary file for lexicon");
      if(!kaldi::ReadLexiconForWordAlign(ki.Stream(), &lexicon)){
        KALDI_ERR << "Error reading alignment lexicon from " << filename_align_lexicon;
      }
    }
    kaldi::WordAlignLatticeLexiconInfo lexicon_info(lexicon);
    { std::vector<std::vector<kaldi::int32> > temp; lexicon.swap(temp); }

    ////////////////////////////////////////////////////////////
    //////////     Configs ()     //////////
    ////////////////////////////////////////////////////////////
    feature_info = new kaldi::OnlineNnet2FeaturePipelineInfo(feature_config);

    //====================================================================================================//
    if(nnet_type == 1){
      ////////////////////////////////////////////////////////////
      //////////          NNET (Acoustic Model)         //////////
      ////////////////////////////////////////////////////////////
      // feature_transform + nnet
      kaldi::nnet1::Nnet nnet;
      {
        kaldi::nnet1::Nnet nnet_am;
        nnet_am.Read(filename_nnet);
        nnet.Read(filename_feature_transform);
        nnet.RemoveComponent(0);
        nnet_am.RemoveLastComponent();
        nnet.AppendNnet(nnet_am);
      }
      std::cout << nnet.Info();

      ////////////////////////////////////////////////////////////
      //////////      PDF Prior (Class Frame Count)     //////////
      ////////////////////////////////////////////////////////////
      kaldi::nnet1::PdfPriorOptions pdf_prior_opts;
      pdf_prior_opts.class_frame_counts = filename_class_frame_counts;
      kaldi::nnet1::PdfPrior pdf_prior(pdf_prior_opts);

      ////////////////////////////////////////////////////////////
      //////////    Transition Model (Acoustic Model)   //////////
      ////////////////////////////////////////////////////////////
      kaldi::TransitionModel trans_model;
      kaldi::ReadKaldiObject(filename_mdl, &trans_model);

      ///////// File List Input (Offline)
      if(!filename_features_list.empty()){
        kaldi::Input input(filename_features_list);
        //kaldi::Timer timer;

        std::map<std::uint32_t, std::thread*> thread_map; // current jobs
        std::uint32_t id = 0;
        while(true){
          for(std::map<std::uint32_t, std::thread*>::iterator it = thread_map.begin(); it != thread_map.end();){
            if(it->second->joinable()){
              std::uint32_t joinable_id = it->first;
              if(rs->find_recogend_key(joinable_id)){
                it->second->join();
                //ms->erase_buffer_key(joinable_id);
                //ms->erase_terminated_key(joinable_id);
                rs->erase_recogend_key(joinable_id);
                delete it->second;
                thread_map.erase(it++);
              }
              else{ ++it; }
            }
            else{ ++it; }
            //DEBUG//KALDI_WARN << "terminate thread_map.size()=" << thread_map.size();
          }
          //DEBUG//KALDI_WARN << "thread_map.size()=" << thread_map.size();

          if(!is_active || input.Stream().eof()){
            if(thread_map.size()==0){ break; }
            multiplatform::time::ms_sleep(100);
            continue;
          }

          if(max_tasks==0 || max_tasks>thread_map.size()){
            std::string filename;
            input.Stream() >> filename;
            if(!multiplatform::file::file_exists(filename)){ continue; }
            //std::cout << "file: " << filename << std::endl;
            _raw_print_with_endl(std::string("file: ").append(filename));

            id++;
            kaldi::Matrix<kaldi::BaseFloat> feat;
            kaldi::HtkHeader header;
            kaldi::Input ki(filename, NULL);
            kaldi::ReadHtk(ki.Stream(), &feat, &header);

            kaldi::OnlineMatrixFeature online_mat_feat(feat);
            //DEBUG//KALDI_LOG << online_mat_feat.Dim() << ", " << online_mat_feat.NumFramesReady();
            std::shared_ptr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(feature_info, feat, id));

            //DEBUG//KALDI_WARN << "recog start id=" << id;
            kaldi::OnlineSpliceOptions splice_opts;
            splice_opts.left_context = splice;
            splice_opts.right_context = splice;
            std::shared_ptr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));
            //DEBUG//KALDI_LOG << splice_frames->Dim() << ", " << splice_frames->NumFramesReady();
            kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt;
            opt.SetAcousticScale(decodable_opts.acoustic_scale);
            kaldi::nnet1::NnetRecognizer* recognizer = // kaldi::nnet1 namespace
              new kaldi::nnet1::NnetRecognizer(nnet, pdf_prior, trans_model, opt, features, splice_frames,
                                                    decode_fst, decoder_config, phone_sym_table, word_sym_table,
                                                    lexicon_info, lm_name, 1, nbest_count, progress_output_interval, rs);
            thread_map[id] = new std::thread(nnet1_recog_proc_thread, recognizer);
          }
          multiplatform::time::ms_sleep(10);
        }
      }

      ///////// SpeechRecognition[SMN]Client (Online)
      else{ // else if(filename_features_list.empty()){
        std::map<std::uint32_t, std::thread*> thread_map; // current jobs
        std::uint32_t id = 0;
        while(true){
          for(std::map<std::uint32_t, std::thread*>::iterator it = thread_map.begin(); it != thread_map.end();){
            if(it->second->joinable()){
              std::uint32_t joinable_id = it->first;
              if(rs->find_recogend_key(joinable_id)){
                it->second->join();
                //ms->erase_buffer_key(joinable_id);
                ms->erase_terminated_key(joinable_id);
                rs->erase_recogend_key(joinable_id);
                delete it->second;
                thread_map.erase(it++);
              }
              else{ ++it; }
            }
            else{ ++it; }
            //DEBUG//KALDI_WARN << "terminate thread_map.size()=" << thread_map.size();
          }
          //DEBUG//KALDI_WARN << "thread_map.size()=" << thread_map.size();

          if(!is_active){
            if(thread_map.size()==0){ break; }
            multiplatform::time::ms_sleep(100);
            continue;
          }

          if(max_tasks==0 || max_tasks>thread_map.size()){
            std::deque<std::uint32_t> new_keys = ms->get_new_keys();
            std::uint32_t free_tasks = (max_tasks==0) ? new_keys.size() : max_tasks - thread_map.size();
            while(is_active && free_tasks>0 && !new_keys.empty()){
              id = new_keys.front(); new_keys.pop_front();
              ms->erase_new_key(id);
              //DEBUG//KALDI_WARN << "new id=" << id;
              std::shared_ptr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(feature_info, ms, id));
              //DEBUG//features->Initialize();
              //DEBUG//id = features->GetID();
              //DEBUG//KALDI_WARN << "recog start id=" << id;
              kaldi::OnlineSpliceOptions splice_opts;
              splice_opts.left_context = splice;
              splice_opts.right_context = splice;
              std::shared_ptr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));
              kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt;
              opt.SetAcousticScale(decodable_opts.acoustic_scale);
              kaldi::nnet1::NnetRecognizer* recognizer = // kaldi::nnet1 namespace
                new kaldi::nnet1::NnetRecognizer(nnet, pdf_prior, trans_model, opt, features, splice_frames,
                                                 decode_fst, decoder_config, phone_sym_table, word_sym_table,
                                                 lexicon_info, lm_name, 1, nbest_count, progress_output_interval, rs);
              thread_map[id] = new std::thread(nnet1_recog_proc_thread, recognizer);
              free_tasks--;
            }
          }
          multiplatform::time::ms_sleep(10);
        }
      } // condition (filename_features_list.empty()) end
    }
    //====================================================================================================//
/*
    else if(nnet_type == 2){
      ////////////////////////////////////////////////////////////
      //////////          NNET (Acoustic Model)         //////////
      ////////////////////////////////////////////////////////////
      kaldi::TransitionModel trans_model;
      kaldi::nnet2::AmNnet am_nnet;
      {
        bool binary;
        kaldi::Input ki(filename_mdl, &binary);
        trans_model.Read(ki.Stream(), binary);
        am_nnet.Read(ki.Stream(), binary);
      }

      ///////// File List Input (Offline)
      if(!filename_features_list.empty()){
        kaldi::Input input(filename_features_list);
        //kaldi::Timer timer;

        std::map<std::uint32_t, std::thread*> thread_map; // current jobs
        std::uint32_t id = 0;
        while(true){
          for(std::map<std::uint32_t, std::thread*>::iterator it = thread_map.begin(); it != thread_map.end();){
            if(it->second->joinable()){
              std::uint32_t joinable_id = it->first;
              if(rs->find_recogend_key(joinable_id)){
                it->second->join();
                //ms->erase_buffer_key(joinable_id);
                //ms->erase_terminated_key(joinable_id);
                rs->erase_recogend_key(joinable_id);
                delete it->second;
                thread_map.erase(it++);
              }
              else{ ++it; }
            }
            else{ ++it; }
            //DEBUG//KALDI_WARN << "terminate thread_map.size()=" << thread_map.size();
          }
          //DEBUG//KALDI_WARN << "thread_map.size()=" << thread_map.size();

          if(!is_active || input.Stream().eof()){
            if(thread_map.size()==0){ break; }
            multiplatform::time::ms_sleep(100);
            continue;
          }

          if(max_tasks==0 || max_tasks>thread_map.size()){
            std::string filename;
            input.Stream() >> filename;
            if(!multiplatform::file::file_exists(filename)){ continue; }
            _raw_print_with_endl(std::string("file: ").append(filename));

            id++;
            kaldi::Matrix<kaldi::BaseFloat> feat;
            kaldi::HtkHeader header;
            kaldi::Input ki(filename, NULL);
            kaldi::ReadHtk(ki.Stream(), &feat, &header);

            kaldi::OnlineMatrixFeature online_mat_feat(feat);
            KALDI_LOG << online_mat_feat.Dim() << ", " << online_mat_feat.NumFramesReady();
            std::shared_ptr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(feature_info, feat, id));

            //DEBUG//KALDI_WARN << "recog start id=" << id;
            kaldi::OnlineSpliceOptions splice_opts;
            splice_opts.left_context = splice;
            splice_opts.right_context = splice;
            std::shared_ptr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));
            KALDI_LOG << splice_frames->Dim() << ", " << splice_frames->NumFramesReady();
            kaldi::nnet2::DecodableNnet2OnlineForHarkOptions opt; opt.SetAcousticScale(ac_scale);
            //kaldi::nnet2::NnetRecognizer* recognizer = // kaldi::nnet2 namespace
            //  new kaldi::nnet2::NnetRecognizer(am_nnet, pdf_prior, trans_model, opt, features, splice_frames,
            //                                   decode_fst, decoder_config, phone_sym_table, word_sym_table,
            //                                   lexicon_info, lm_name, 1, nbest_count, progress_output_interval, rs);
            //thread_map[id] = new std::thread(nnet2_recog_proc_thread, recognizer);
          }
          multiplatform::time::ms_sleep(10);
        }
      }

      ///////// HarkDataStreamReceiver (Online)
      else{ // else if(filename_features_list.empty()){
        std::map<std::uint32_t, std::thread*> thread_map; // current jobs
        std::uint32_t id = 0;
        while(true){
          for(std::map<std::uint32_t, std::thread*>::iterator it = thread_map.begin(); it != thread_map.end();){
            if(it->second->joinable()){
              std::uint32_t joinable_id = it->first;
              if(rs->find_recogend_key(joinable_id)){
                it->second->join();
                //ms->erase_buffer_key(joinable_id);
                ms->erase_terminated_key(joinable_id);
                rs->erase_recogend_key(joinable_id);
                delete it->second;
                thread_map.erase(it++);
              }
              else{ ++it; }
            }
            else{ ++it; }
            //DEBUG//KALDI_WARN << "terminate thread_map.size()=" << thread_map.size();
          }
          //DEBUG//KALDI_WARN << "thread_map.size()=" << thread_map.size();

          if(!is_active){
            if(thread_map.size()==0){ break; }
            multiplatform::time::ms_sleep(100);
            continue;
          }

          if(max_tasks==0 || max_tasks>thread_map.size()){
            std::deque<std::uint32_t> new_keys = ms->get_new_keys();
            std::uint32_t free_tasks = (max_tasks==0) ? new_keys.size() : max_tasks - thread_map.size();
            while(is_active && free_tasks>0 && !new_keys.empty()){
              id = new_keys.front(); new_keys.pop_front();
              ms->erase_new_key(id);
              //DEBUG//KALDI_WARN << "new id=" << id;
              std::shared_ptr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(feature_info, ms, id));
              //DEBUG//features->Initialize();
              //DEBUG//id = features->GetID();
              //DEBUG//KALDI_WARN << "recog start id=" << id;
              kaldi::OnlineSpliceOptions splice_opts;
              splice_opts.left_context = splice;
              splice_opts.right_context = splice;
              std::shared_ptr<kaldi::OnlineFeatureInterface> splice_frames(new kaldi::OnlineSpliceFrames(splice_opts, features.get()));
              kaldi::nnet2::DecodableNnet2OnlineForHarkOptions opt; opt.SetAcousticScale(ac_scale);
              //kaldi::nnet2::NnetRecognizer* recognizer = // kaldi::nnet2 namespace
              //  new kaldi::nnet2::NnetRecognizer(am_nnet, pdf_prior, trans_model, opt, features, splice_frames,
              //                                   decode_fst, decoder_config, phone_sym_table, word_sym_table,
              //                                   lexicon_info, lm_name, 1, nbest_count, progress_output_interval, rs);
              //thread_map[id] = new std::thread(nnet2_recog_proc_thread, recognizer);
              free_tasks--;
            }
          }
          multiplatform::time::ms_sleep(10);
        }
      } // condition (filename_features_list.empty()) end
    }
*/
    //====================================================================================================//
    else if(nnet_type == 3){
      //KALDI_ERR << "Sorry nnet3 model is not support yet.";

      ////////////////////////////////////////////////////////////
      //////////          NNET (Acoustic Model)         //////////
      ////////////////////////////////////////////////////////////
      kaldi::TransitionModel trans_model;
      kaldi::nnet3::AmNnetSimple am_nnet;
      {
        bool binary;
        kaldi::Input ki(filename_mdl, &binary);
        trans_model.Read(ki.Stream(), binary);
        am_nnet.Read(ki.Stream(), binary);
        kaldi::nnet3::SetBatchnormTestMode(true, &(am_nnet.GetNnet()));
        kaldi::nnet3::SetDropoutTestMode(true, &(am_nnet.GetNnet()));
        kaldi::nnet3::CollapseModel(kaldi::nnet3::CollapseModelConfig(), &(am_nnet.GetNnet()));
      }
      kaldi::nnet3::DecodableNnetSimpleLoopedInfo decodable_info(decodable_opts, &am_nnet);

      ///////// File List Input (Offline)
      if(!filename_features_list.empty()){
        kaldi::Input input(filename_features_list);
        //kaldi::Timer timer;

        std::map<std::uint32_t, std::thread*> thread_map; // current jobs
        std::uint32_t id = 0;
        while(true){
          for(std::map<std::uint32_t, std::thread*>::iterator it = thread_map.begin(); it != thread_map.end();){
            if(it->second->joinable()){
              std::uint32_t joinable_id = it->first;
              if(rs->find_recogend_key(joinable_id)){
                it->second->join();
                //ms->erase_buffer_key(joinable_id);
                //ms->erase_terminated_key(joinable_id);
                rs->erase_recogend_key(joinable_id);
                delete it->second;
                thread_map.erase(it++);
              }
              else{ ++it; }
            }
            else{ ++it; }
            //DEBUG//KALDI_WARN << "terminate thread_map.size()=" << thread_map.size();
          }
          //DEBUG//KALDI_WARN << "thread_map.size()=" << thread_map.size();

          if(!is_active || input.Stream().eof()){
            if(thread_map.size()==0){ break; }
            multiplatform::time::ms_sleep(100);
            continue;
          }

          if(max_tasks==0 || max_tasks>thread_map.size()){
            std::string filename;
            input.Stream() >> filename;
            if(!multiplatform::file::file_exists(filename)){ continue; }
            _raw_print_with_endl(std::string("file: ").append(filename));

            id++;
            kaldi::Matrix<kaldi::BaseFloat> feat;
            kaldi::HtkHeader header;
            kaldi::Input ki(filename, NULL);
            kaldi::ReadHtk(ki.Stream(), &feat, &header);

            kaldi::OnlineMatrixFeature online_mat_feat(feat);
            //DEBUG//KALDI_LOG << online_mat_feat.Dim() << ", " << online_mat_feat.NumFramesReady();
            std::shared_ptr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(feature_info, feat, id));
            //DEBUG//KALDI_WARN << "recog start id=" << id;
            std::shared_ptr<kaldi::OnlineFeatureInterface> features_if(features);
            //DEBUG//KALDI_LOG << features_if->Dim() << ", " << features_if->NumFramesReady();
            kaldi::nnet3::NnetRecognizer* recognizer = // kaldi::nnet3 namespace
              new kaldi::nnet3::NnetRecognizer(am_nnet, trans_model, decodable_opts, decodable_info, feature_info,
                                               features, features_if,
                                               decode_fst, decoder_config, phone_sym_table, word_sym_table,
                                               lexicon_info, lm_name, 1, nbest_count, progress_output_interval, rs);
            thread_map[id] = new std::thread(nnet3_recog_proc_thread, recognizer);
          }
          multiplatform::time::ms_sleep(10);
        }
      }

      ///////// HarkDataStreamReceiver (Online)
      else{ // else if(filename_features_list.empty()){
        std::map<std::uint32_t, std::thread*> thread_map;
        std::uint32_t id = 0;
        while(true){
          for(std::map<std::uint32_t, std::thread*>::iterator it = thread_map.begin(); it != thread_map.end();){
            if(it->second->joinable()){
              std::uint32_t joinable_id = it->first;
              if(rs->find_recogend_key(joinable_id)){
                it->second->join();
                //ms->erase_buffer_key(joinable_id);
                ms->erase_terminated_key(joinable_id);
                rs->erase_recogend_key(joinable_id);
                delete it->second;
                thread_map.erase(it++);
              }
              else{
                if(!is_active){
                  // To raise the shutdown response speed, a thread whose output of the recognition results
                  // are not completed in time relinquishes its management.
                  it->second->detach();
                  delete it->second;
                  thread_map.erase(it++);
                }
                else{ ++it; }
              }
            }
            else{ ++it; }
            //DEBUG//KALDI_WARN << "terminate thread_map.size()=" << thread_map.size();
          }
          //DEBUG//KALDI_WARN << "thread_map.size()=" << thread_map.size();

          if(!is_active){
            if(thread_map.size()==0){ break; }
            multiplatform::time::ms_sleep(100);
            continue;
          }

          if(max_tasks==0 || max_tasks>thread_map.size()){
            std::deque<std::uint32_t> new_keys = ms->get_new_keys();
            std::uint32_t free_tasks = (max_tasks==0) ? new_keys.size() : max_tasks - thread_map.size();
            while(is_active && free_tasks>0 && !new_keys.empty()){
              id = new_keys.front(); new_keys.pop_front();
              ms->erase_new_key(id);
              //DEBUG//KALDI_WARN << "new id=" << id;
              std::shared_ptr<kaldi::OnlineFeatureMfcnet> features(new kaldi::OnlineFeatureMfcnet(feature_info, ms, id));
              //DEBUG//features->Initialize();
              //DEBUG//id = features->GetID();
              //DEBUG//KALDI_WARN << "recog start id=" << id;

              std::shared_ptr<kaldi::OnlineFeatureInterface> features_if(features);
              //DEBUG//KALDI_LOG << features_if->Dim() << ", " << features_if->NumFramesReady();

              kaldi::nnet3::NnetRecognizer* recognizer = // kaldi::nnet3 namespace
                new kaldi::nnet3::NnetRecognizer(am_nnet, trans_model, decodable_opts, decodable_info, feature_info,
                                                 features, features_if,
                                                 decode_fst, decoder_config, phone_sym_table, word_sym_table,
                                                 lexicon_info, lm_name, 1, nbest_count, progress_output_interval, rs);
              thread_map[id] = new std::thread(nnet3_recog_proc_thread, recognizer);
              free_tasks--;
            }
          }
          multiplatform::time::ms_sleep(10);
        }
      } // condition (filename_features_list.empty()) end
    }
    //====================================================================================================//
    else{
      KALDI_ERR << "Could not support nnet type nnet" << nnet_type << ".";
    }
    //====================================================================================================//

    ////////////////////////////////////////////////////////////
    //////////         Finalize (Free Memory)         //////////
    ////////////////////////////////////////////////////////////
    if(decode_fst != nullptr){ delete decode_fst; }
    if(phone_sym_table != nullptr){ delete phone_sym_table; }
    if(word_sym_table != nullptr){ delete word_sym_table; }
    if(feature_info != nullptr){ delete feature_info; }
  } // common scope end

  return 0;
}
