
#ifndef MP_RESULT_SERVER_H_
#define MP_RESULT_SERVER_H_

#include "hark-support.h"
//DEBUG//#include "stdio_synchronizer.hpp"
#include "io_manager.hpp"
#include "tcp_server.hpp"
#include "tcp_socket.hpp"
#include "thread_pool.hpp"
#include "wsa_wrapper.hpp"

//DEBUG//#include <sstream>
#include <iostream>
#include <mutex>
//#include <map>
#include <unordered_map>
#include <deque>
#include <utility>
#include <functional>
#include <cstdint>

class result_server {
public:
  result_server(){
  }

  ~result_server(){
    std::lock_guard<std::mutex> lock(m_mutex_buffers);
    buffers.erase(buffers.begin(), buffers.end());
  }

  result_server(const std::string& host, std::uint32_t port){
    s.start(host, port, [&](const std::shared_ptr<multiplatform::tcp_client>& client) -> bool {
      //DEBUG// std::stringstream ss;
      //DEBUG// ss << "New client: " << client->get_port();
      //DEBUG// raw_print_with_endl(ss.str());
      {
        std::lock_guard<std::mutex> lock(m_mutex_buffers);
        buffers[client->get_port()] = client;
      }
      client->async_read({1024, std::bind(&result_server::on_new_message, this, client, std::placeholders::_1)});
      return true;
    });
  }

  std::size_t broadcast(std::string text){
    std::lock_guard<std::mutex> lock(m_mutex_buffers);
    std::vector<char> vec(text.begin(), text.end());
    for(auto buffer : buffers){
      if(buffer.second->is_connected()){
        buffer.second->async_write({vec, nullptr});
      }
    }
    return 0;
  }

  bool add_recogend_key(std::uint32_t& port){
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_emap_mutex);
      auto p = emap.insert(std::pair<std::uint32_t, bool>(port, true));
      ret = p.second;
    }
    return ret;
  }

  bool erase_recogend_key(std::uint32_t& port){
    if(emap.empty()){ return false; }
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_emap_mutex);
      if(emap.erase(port) != 0){ ret = true; }
    }
    return ret;
  }

  bool find_recogend_key(std::uint32_t& port){
    if(emap.empty()){ return false; }
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_emap_mutex);
      if(emap.find(port) != emap.end()){ ret = true; }
    }
    return ret;
  }

  void on_new_message(const std::shared_ptr<multiplatform::tcp_client>& client, multiplatform::tcp_client::read_response& res) {
    if(res.success){
      //DEBUG// raw_print_with_endl("Client recv data");
      client->async_read({1024, std::bind(&result_server::on_new_message, this, client, std::placeholders::_1)});
    }
    else {
      //DEBUG// raw_print_with_endl("Client disconnected");
      client->disconnect();
      {
        std::lock_guard<std::mutex> lock(m_mutex_buffers);
        buffers.erase(client->get_port());
      }
    }
  }

  multiplatform::tcp_server s;
  std::unordered_map<std::uint32_t, std::shared_ptr<multiplatform::tcp_client> > buffers; // key:port, data:buffer
  std::unordered_map<std::uint32_t, bool> emap; // recognition end list (recongnition complete sources)
  std::mutex m_mutex_buffers;
  std::mutex m_emap_mutex;
};

#endif // MP_RESULT_SERVER_H_

